-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}


local gearbox = nil
local engine = nil
local motors = nil

local behavior = nil
local sharedFunctions = nil

local masterShiftLogic = nil
local slaveShiftLogic = nil
local mainShiftLogic = nil
local hasMasterShiftLogic = false

M.gearboxHandling = nil
M.timer = nil
M.timerConstants = nil
M.inputValues = nil
M.shiftPreventionData = nil
M.shiftBehavior = nil
M.smoothedValues = nil

M.currentGearIndex = 0
M.maxGearIndex = 0
M.minGearIndex = 0
M.throttle = 0
M.brake = 0
M.regen = 0
M.clutchRatio = 0
M.shiftingAggression = 0
M.isArcadeSwitched = false
M.isSportModeActive = false
M.isShifting = false

M.smoothedAvgAVInput = 0
M.rpm = 0
M.idleRPM = 0
M.maxRPM = 0

M.engineThrottle = 0
M.engineLoad = 0
M.engineTorque = 0
M.flywheelTorque = 0
M.gearboxTorque = 0

M.ignition = true
M.isEngineRunning = 0

M.oilTemp = 0
M.waterTemp = 0
M.checkEngine = false

M.energyStorages = {}

local function syncGear(offset)
  if behavior ~= "realistic" then
    return
  end
  if hasMasterShiftLogic then -- sync gear
    local offset = offset or 0
    local goalGear = 2
    local _, index = masterShiftLogic.getGearPosition()
    if index == 0 then -- numeric gear index
      local predictGear = masterShiftLogic.currentGearIndex + offset
      goalGear = predictGear > 0 and 2 or (predictGear < 0 and -1) or 0
    else -- automatic gear index (PRNDS21M like)
      goalGear = index == 1 and 1 or (index < 4 and index - 3) or 2
    end
    slaveShiftLogic.shiftToGearIndex(goalGear)
  end
end

local function shiftUpOnDown()
  syncGear(1)

  if mainShiftLogic.shiftUpOnDown then
    mainShiftLogic.shiftUpOnDown()
  elseif mainShiftLogic.shiftUp then
    mainShiftLogic.shiftUp()
  end
end
local function shiftUpOnUp()
  if mainShiftLogic.shiftUpOnUp then
    mainShiftLogic.shiftUpOnUp()
  end
end

local function shiftDownOnDown()
  syncGear(-1)

  if mainShiftLogic.shiftDownOnDown then
    mainShiftLogic.shiftDownOnDown()
  elseif mainShiftLogic.shiftDown then
    mainShiftLogic.shiftDown()
  end
end

local function shiftDownOnUp()
  if mainShiftLogic.shiftDownOnUp then
    mainShiftLogic.shiftDownOnUp()
  end
end

local function shiftToGearIndex(index)
  mainShiftLogic.shiftToGearIndex(index)
  if hasMasterShiftLogic then -- sync gear
    slaveShiftLogic.shiftToGearIndex(index)
  end
end

local lastIgnitionLevel = nil
local function updateGearboxGFX(dt)
  -- syncing
  mainShiftLogic.gearboxHandling =     M.gearboxHandling -- READ ONLY
  mainShiftLogic.timer =               M.timer -- READ & WRITE
  mainShiftLogic.timerConstants =      M.timerConstants -- READ ONLY
  mainShiftLogic.inputValues =         M.inputValues -- READ ONLY
  mainShiftLogic.shiftPreventionData = M.shiftPreventionData -- READ ONLY
  mainShiftLogic.shiftBehavior =       M.shiftBehavior -- READ ONLY
  mainShiftLogic.smoothedValues =      M.smoothedValues -- READ ONLY
  if hasMasterShiftLogic then
    slaveShiftLogic.gearboxHandling =      M.gearboxHandling -- READ ONLY
    slaveShiftLogic.timer =                M.timer -- READ & WRITE
    slaveShiftLogic.timerConstants =       M.timerConstants -- READ ONLY
    slaveShiftLogic.inputValues =          M.inputValues -- READ ONLY
    slaveShiftLogic.shiftPreventionData =  M.shiftPreventionData -- READ ONLY
    slaveShiftLogic.shiftBehavior =        M.shiftBehavior -- READ ONLY
    slaveShiftLogic.smoothedValues =       M.smoothedValues -- READ ONLY
  end

  -- wrapped calls
  slaveShiftLogic.updateGearboxGFX(dt)
  if masterShiftLogic then
    masterShiftLogic.updateGearboxGFX(dt)

    local ignitionLevel = electrics.values.ignitionLevel
    -- sync motor ignition with electrics ignition state
    for _, motor in ipairs(motors) do
      motor:setIgnition(ignitionLevel >= 2 and 1 or 0)
    end

    -- handle ******** auto parking gear logic from emotor shift logic
    --local masterGearName = masterShiftLogic.getGearName()
    --if ignitionLevel <= 1 or ignitionLevel ~= lastIgnitionLevel then
    --  lastIgnitionLevel = ignitionLevel
    --  syncGear()
    --end
  end
  syncGear()

  -- post processing (IM NOT SHADER)

  M.regen =              slaveShiftLogic.regen
  M.currentGearIndex =   mainShiftLogic.currentGearIndex
  M.throttle =           mainShiftLogic.throttle
  M.brake =              mainShiftLogic.brake
  M.clutchRatio =        mainShiftLogic.clutchRatio
  
  M.rpm =                mainShiftLogic.rpm
  M.smoothedAvgAVInput = mainShiftLogic.smoothedAvgAVInput
  M.waterTemp =          mainShiftLogic.waterTemp
  M.oilTemp =            mainShiftLogic.oilTemp
  M.checkEngine =        mainShiftLogic.checkEngine
  M.ignition =           mainShiftLogic.ignition
  M.engineThrottle =     mainShiftLogic.engineThrottle
  M.engineLoad =         mainShiftLogic.engineLoad
  M.running =            mainShiftLogic.running
  M.engineTorque =       mainShiftLogic.engineTorque
  M.flywheelTorque =     mainShiftLogic.flywheelTorque
  M.gearboxTorque =      mainShiftLogic.gearboxTorque
  M.isEngineRunning =    mainShiftLogic.isEngineRunning
  M.isShifting =         mainShiftLogic.isShifting
  M.shiftingAggression = mainShiftLogic.shiftingAggression
  M.isArcadeSwitched =   mainShiftLogic.isArcadeSwitched
  M.isSportModeActive =  mainShiftLogic.isSportModeActive

  M.shiftUp =            mainShiftLogic.shiftUp
  M.shiftDown =          mainShiftLogic.shiftDown
end

local function sendTorqueData()
  if hasMasterShiftLogic then
    masterShiftLogic.sendTorqueData()
  end
  slaveShiftLogic.sendTorqueData()
end

local function init(jbeamData, sharedFunctionTable)
  sharedFunctions = sharedFunctionTable

  motors = {}
  local motorNames = jbeamData.motorNames or {"mainMotor"}
  for _, v in ipairs(motorNames) do
    local motor = powertrain.getDevice(v)
    if motor then
      table.insert(motors, motor)
    end
  end
  engine = powertrain.getDevice("mainEngine")
  gearbox = powertrain.getDevice("gearbox")

  hasMasterShiftLogic = gearbox ~= nil
  if hasMasterShiftLogic then
    masterShiftLogic = require("controller/shiftLogic-"..gearbox.type)
    masterShiftLogic.init(jbeamData, sharedFunctions)
  end
  slaveShiftLogic = require("controller/shiftLogic-electricMotor")
  slaveShiftLogic.init(jbeamData, sharedFunctions)

  mainShiftLogic = slaveShiftLogic
  if hasMasterShiftLogic then
    mainShiftLogic = masterShiftLogic
  end

  -- post process (IM NOT SHADER)
  M.regen =            slaveShiftLogic.regen
  M.currentGearIndex = mainShiftLogic.currentGearIndex
  M.throttle =         mainShiftLogic.throttle
  M.brake =            mainShiftLogic.brake
  M.clutchRatio =      mainShiftLogic.clutchRatio

  M.maxRPM =           mainShiftLogic.maxRPM
  M.idleRPM =          mainShiftLogic.idleRPM
  M.maxGearIndex =     mainShiftLogic.maxGearIndex
  M.minGearIndex =     mainShiftLogic.minGearIndex
  local energyStorages = {}
  for _, energyStorage in ipairs(slaveShiftLogic.energyStorages) do
    table.insert(energyStorages, energyStorage)
  end
  if hasMasterShiftLogic then
    for _, energyStorage in ipairs(masterShiftLogic.energyStorages) do
      table.insert(energyStorages, energyStorage)
    end
  end
  M.energyStorages = energyStorages

  -- end of file direct func linking but done here cause the file are require()ed on init instead of script run
  M.shiftUp =                mainShiftLogic.shiftUp
  M.shiftDown =              mainShiftLogic.shiftDown
  M.getGearName =            mainShiftLogic.getGearName
  M.getGearPosition =        mainShiftLogic.getGearPosition
  M.setIgnition =            slaveShiftLogic.setIgnition -- NOT CALLED WHEN I-C-E, NEED TO SYNC THIS OURSELF

  M.getState = mainShiftLogic.getState -- do i wrap this?
  M.setState = mainShiftLogic.setState -- do i wrap this?
end

local function gearboxBehaviorChanged(Behavior)
  behavior = Behavior
  slaveShiftLogic.gearboxHandling = M.gearboxHandling
  slaveShiftLogic.gearboxBehaviorChanged(behavior)
  if masterShiftLogic then
    masterShiftLogic.gearboxHandling = M.gearboxHandling
    masterShiftLogic.gearboxBehaviorChanged(behavior)
  end
end

local function onDeserialize(data)
  if hasMasterShiftLogic then
    if masterShiftLogic.onDeserialize then
      masterShiftLogic.onDeserialize(data.master)
    end
  end
  if slaveShiftLogic.onDeserialize then
    slaveShiftLogic.onDeserialize(data.slave)
  end
end

local function onSerialize()
  local data = {}
  if hasMasterShiftLogic then
    data.master = masterShiftLogic.onSerialize and masterShiftLogic.onSerialize()
  end
  data.slave = slaveShiftLogic.onSerialize and slaveShiftLogic.onSerialize()
  return data
end

M.init = init

 -- CUSTOM WRAPPER
M.updateGearboxGFX = updateGearboxGFX
M.sendTorqueData = sendTorqueData
M.gearboxBehaviorChanged = gearboxBehaviorChanged -- warpped to fix issue with bad init values
M.shiftUpOnDown = shiftUpOnDown
M.shiftDownOnDown = shiftDownOnDown
M.shiftUpOnUp = shiftUpOnUp
M.shiftDownOnUp = shiftDownOnUp
M.onDeserialize = onDeserialize
M.onSerialize = onSerialize

return M
